"""A module designed for optimizing the combined ligand & core.

Index
-----
.. currentmodule:: CAT.attachment.qd_opt
.. autosummary::
    init_qd_opt
    start_qd_opt
    get_job_settings
    _qd_to_db
    qd_opt

API
---
.. autofunction:: init_qd_opt
.. autofunction:: start_qd_opt
.. autofunction:: get_job_settings
.. autofunction:: _qd_to_db
.. autofunction:: qd_opt

"""

from typing import Tuple, Iterable, Optional, Type, NoReturn, Any

from scm.plams import Molecule, Settings, AMSJob
from scm.plams.core.basejob import Job

from ..jobs import job_geometry_opt  # noqa: F401
from ..workflows import WorkFlow, MOL, JOB_SETTINGS_QD_OPT
from ..mol_utils import fix_carboxyl, fix_h
from ..settings_dataframe import SettingsDataFrame
from ..data_handling.mol_to_file import mol_to_file

try:
    from nanoCAT.qd_opt_ff import qd_opt_ff
except ImportError as ex:
    def qd_opt_ff(*args: Any, ex: Exception = ex, **kwargs: Any) -> NoReturn: raise ex

__all__ = ['init_qd_opt']


def init_qd_opt(qd_df: SettingsDataFrame) -> None:
    """Initialize the ligand optimization procedure."""
    workflow = WorkFlow.from_template(qd_df, name='qd_opt')

    # Pull from the database; push unoptimized structures
    idx = workflow.from_db(qd_df)
    workflow(start_qd_opt, qd_df, columns=[], index=idx)

    # Sets a nested list
    # This cannot be done with loc is it will try to expand the list into a 2D array
    qd_df[JOB_SETTINGS_QD_OPT] = workflow.pop_job_settings(qd_df[MOL])

    # Push the optimized structures to the database
    job_recipe = workflow.get_recipe()
    workflow.to_db(qd_df, status='optimized', index=idx, job_recipe=job_recipe)

    # Export ligands to .xyz, .pdb, .mol and/or .mol format
    mol_format = qd_df.settings.optional.database.mol_format
    if mol_format:
        path = workflow.path
        mol_to_file(qd_df.loc[idx, MOL], path, mol_format=mol_format)


def start_qd_opt(mol_list: Iterable[Molecule],
                 jobs: Tuple[Optional[Type[Job]], ...], settings: Tuple[Optional[Settings], ...],
                 use_ff: bool = False, **kwargs) -> None:
    """Loop over all molecules in **mol_list** and perform geometry optimizations."""
    for mol in mol_list:
        mol.properties.job_path = []
        qd_opt(mol, jobs, settings, use_ff=use_ff)


def qd_opt(mol: Molecule, jobs: Tuple[Optional[Type[Job]], ...],
           settings: Tuple[Optional[Settings], ...], use_ff: bool = False) -> None:
    """Perform an optimization of the quantum dot.

    Performs an inplace update of **mol**.

    Parameters
    ----------
    mol : |plams.Molecule|_
        The to-be optimized molecule.

    job_recipe : |plams.Settings|_
        A Settings instance containing all jon settings.
        Expects 4 keys: ``"job1"``, ``"job2"``, ``"s1"``, ``"s2"``.

    forcefield : bool
        If ``True``, perform the job with CP2K with a user-specified forcefield.

    """
    # Prepare the job settings
    if use_ff:
        qd_opt_ff(mol, jobs, settings)
        return None

    # Expand arguments
    job1, job2 = jobs
    s1, s2 = settings

    # Extra options for AMSJob
    if job1 is AMSJob:
        s1 = Settings(s1)
        s1.input.ams.constraints.atom = mol.properties.indices
    if job2 is AMSJob:
        s2 = Settings(s2)
        s2.input.ams.constraints.atom = mol.properties.indices

    # Run the first job and fix broken angles
    mol.job_geometry_opt(job1, s1, name='QD_opt_part1')
    fix_carboxyl(mol)
    fix_h(mol)
    mol.round_coords()

    # Run the second job
    if job2 is not None:
        mol.job_geometry_opt(job2, s2, name='QD_opt_part2')
        mol.round_coords()
    return None



"""A modules for combining (organic) molecules."""

from itertools import chain

import numpy as np
from scipy.spatial.distance import cdist

from rdkit.Chem import AllChem
from scm.plams import Molecule, Settings, to_rdmol

from .ligand_attach import rot_mol
from ..logger import logger

__all__ = ['label_lig', 'label_core', 'substitution', 'multi_substitution']

# flake8: noqa: N806


def ff_constrained_opt(mol, constrain=()):
    """Performs a constrained FF optimization on a PLAMS molecule.

    Optimisation with rdkit.Chem.AllChem.MMFFGetMoleculeForceField; a PLAMS molecule
    is converted to rdkit molecule, optimised with frozen atoms and converted back to
    PLAMS molecule.

    Parameters
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule.

    constrain : list
        A list of indices of to-be frozen atoms.

    Returns
    -------
    |plams.Molecule|
        Optimized molecular structure

    """
    rdkit_mol = to_rdmol(mol)
    ff_type = AllChem.MMFFGetMoleculeForceField

    try:
        ff = ff_type(rdkit_mol, AllChem.MMFFGetMoleculeProperties(rdkit_mol))
        for f in constrain:
            ff.AddFixedPoint(f)
    except AttributeError:
        # It seems like the MMF forcefield is not available for all atoms (e.g. As)
        logger.warning(
            f"Failed to construct the {ff_type.__name__!r} forcefield for {mol.get_formula()}"
        )
        return mol

    ff.Minimize()
    mol.from_rdmol(rdkit_mol)

    return mol


def connect_ligands_to_core(lig_dict, core, user_min_dist):
    """Attaches multiple ligands to multiple copies of a single core.

    Returns a list of cores with attached ligands, each with the properties.min_distance attribute
    containing the smallest distance between ligand and core.

    Parameters
    ----------
    lig_dict : dict
        Ligands that are attached to the core molecule
        An iterable container consisting of a PLAMS molecules - ligand;
        A np.array of PLAMS atom coordinates - atom to be substituted;
        A vector array - bond between substituent and rest of the compound;
        An int - enumerating ligands
    core : |plams.Molecule|
        A PLAMS molecule - core on which ligands are attached
    user_min_dist : float
        Value for the minimal bond distance in the new molecule

    Returns
    -------
    list of PLAMS molecules
        A list of new molecule that are made core with different ligands attached

    """
    # Unpack the ligand dictionary
    lig_list = lig_dict['lig_list']
    lig_idx = lig_dict['lig_idx']
    lig_vec = lig_dict['lig_vec']
    lig_IDs = lig_dict['lig_ID']

    # Return if core.properties.vec has been exhausted
    try:
        core_vec = core.properties.vec[0]
    except IndexError:
        return []

    # Construct keyword arguments
    kwarg = get_args(core, lig_list, lig_idx)

    # Allign the ligands with the core; perform the rotation check
    lig_array, min_dist_array = rot_mol(lig_list, lig_vec, core_vec, **kwarg)

    # Reading list of ligands that are already attached to the core
    old_ligID = core.properties.get('ligID', '')

    # Combine the rotated ligands and core into new molecules
    ret = []
    for lig, xyz, min_dist, ligID in zip(lig_list, lig_array, min_dist_array, lig_IDs):
        # Copy and manipulate ligand
        lig_cp = lig.copy()
        lig_cp.guess_bonds()
        lig_cp.from_array(xyz)
        lig_cp.properties = Settings()

        # Adding new ligand to the list
        new_ligID = str(old_ligID) + str(ligID)

        # Copy and manipulate core
        core_h = core.properties.the_h
        core_cp = core.copy()
        core_cp.guess_bonds()
        core_cp.delete_atom(core_cp.closest_atom(core_h))
        core_cp.properties.ligID = new_ligID

        # Merge core and ligand
        lig_cp += core_cp
        lig_cp.properties.name = core.properties.name + "_" + lig.properties.name
        lig_cp.properties.min_distance = min_dist

        # Add bond between core and ligand
        the_h = lig_cp.closest_atom(lig_cp.properties.the_h)
        core_other = lig_cp.closest_atom(lig_cp.properties.coords_other_arrays[len(new_ligID)-1])
        lig_cp.add_bond(the_h, core_other)

        # Make list of atom indices to freeze, FrInd_rdkit, for optimization (rdkit starts at 0!)
        # Atom that ligand is conected to is not frozen
        Conn = list(lig_cp).index(core_other)
        FrInd_rdkit = [x for x in range(len(lig_cp)-len(core_cp), len(lig_cp)) if x != Conn]

        # FF for molecule with frozen core
        try:
            lig_cp = ff_constrained_opt(lig_cp, constrain=FrInd_rdkit)
        except ValueError:
            logger.warning("FF optimization error")

        # Distance between core and ligands
        FrInd_plams = (np.array(FrInd_rdkit)+1).tolist()     # plams counts from 1
        Fr_atoms = np.array([lig_cp[c].coords for c in FrInd_plams])
        OnlyLig = np.array(
                [lig_cp[i].coords for i in range(1, len(lig_cp)+1) if i not in FrInd_plams])
        min_dist = np.nanmin(cdist(OnlyLig, Fr_atoms))
        lig_cp.properties.min_distance = min_dist

        ret.append(lig_cp)

    return ret


def get_args(core, lig_list, lig_idx):
    """Extract the various arguments from core and ligand_list.

    Parameters
    ----------
    core : |plams.Molecule|
        A PLAMS molecule with plams_mol.properties attributes
    lig_list : list
        A list with PLAMS molecules with plams_mol.properties attributes
    lig_idx :
        A np.array of PLAMS atom coordinates - atom for substitution

    Returns
    -------
    dict
        A dictionary with properties needed for aligning ligand with the core

    """
    # Extract the various arguments from core and ligand_list
    core_other_atom = core.properties.coords_other_atom[0]
    lig_other = [lig[lig.properties.idx_other+1] for lig in lig_list]
    bond_length = np.array([core_other_atom.radius + lig.radius for lig in lig_other])

    # Roll through the arguments of core
    core.properties.the_h = core.properties.coords_h[0]
    at_h = core.closest_atom(core.properties.the_h)
    core.properties.vec = core.properties.vec[1:]
    core.properties.coords_other_atom = core.properties.coords_other_atom[1:]
    core.properties.coords_other = core.properties.coords_other[1:]
    core.properties.coords_h = core.properties.coords_h[1:]

    # Convert core into an array and mask core_h with np.nan
    core_array = core.as_array()
    idx = core.atoms.index(at_h)
    core_array[idx] = np.nan

    return {'core': core_array,
            'atoms_other': core_other_atom,
            'bond_length': bond_length,
            'dist_to_self': False,
            'idx': lig_idx,
            'ret_min_dist': True}


def label_core(mol):
    """Adds plams_mol.properties attribute to the core.

    Reads the atom indices from comment section in core's .xyz file and adds
    additional plams_mol.properties: coordinates of atom that will be substituted,
    bond vector between the substitution atom and its connection at the core,
    coordinates of the connection at the core

    Parameters
    ----------
    mol : |plams.Molecule|
        An input  PLAMS molecule with atom indices to be substituted in
        plams_mol.properties.comment

    Returns
    -------
    mol : |plams.Molecule|
        A PLAMS mol with additional plams_mol.properties

    """
    # Read the comment in the second line of the xyz file
    comment = mol.properties.comment
    comment = comment.split()
    mol.properties.core_len = len(mol)

    idx = np.array(comment, dtype=int)
    at_h = [mol[int(i)] for i in idx]
    at_other = [at.bonds[0].other_end(at) for at in at_h]

    dummy = Molecule()
    mol.properties.coords_h = dummy.as_array(atom_subset=at_h)
    mol.properties.coords_other = dummy.as_array(atom_subset=at_other)
    mol.properties.vec = mol.properties.coords_other - mol.properties.coords_h

    mol.properties.coords_h_atom = at_h
    mol.properties.coords_other_atom = at_other
    mol.properties.coords_other_arrays = [np.array(i.coords) for i in at_other]
    mol.guess_bonds()


def label_lig(mols):
    """Adds plams_mol.properties attribute to the ligand.

    Reads the atom index from comment section in ligand's .xyz file and adds additional
    plams_mol.properties: coordinates of atom that will be substituted; bond vector between
    the substitution atom and its connection at the ligand; coordinates of the connection
    at the ligand; ligands identity - ligands serial number

    Parameters
    ----------
    mol : |plams.Molecule|
        An input  PLAMS molecule with atom indices to be substituted in plams_mol.properties.comment

    Returns
    -------
    mol : |plams.Molecule|
        A PLAMS mol with additional plams_mol.properties

    """
    ligID = list(range(len(mols)))

    for mol, id in zip(mols, ligID):
        mol.properties.ligID = id
        mol.guess_bonds()

        # Read the comment in the second line of the xyz file
        comment = mol.properties.comment

        # Set an attirbute with indices and vectors
        mol.properties.vec = np.empty(3)

        # Fill the vector array, delete the marked atom in the ligand .xyz file
        at = mol[int(comment)]
        at_other = at.bonds[0].other_end(at)
        mol.properties.vec = np.array(at.coords) - np.array(at_other.coords)

        mol.delete_atom(at)
        mol.properties.idx_other = mol.atoms.index(at_other)
        mol.guess_bonds()


def substitution(input_ligands, input_cores, min_dist):
    """Substitutes atoms at the core with ligands.

    Parameters
    ----------
    input_ligands : list
        A list of input PLAMS molecules with index of an atom to be substituted in
            plams_mol.properties.comment input_cores : list
        A list of input PLAMS molecules with list of indices of atoms to be
            substituted in plams_mol.properties.comment min_dist : float
        Minimal distance between core and attached ligands

    Returns
    -------
    list
        New molecules that are made of ligands attached to the core at position
        of the first index in the list
    """
    lig_idx = np.array([lig.properties.idx_other for lig in input_ligands])
    lig_vec = np.array([lig.properties.vec for lig in input_ligands])
    lig_ID = [lig.properties.ligID for lig in input_ligands]
    lig_dict = {'lig_list': input_ligands, 'lig_idx': lig_idx, 'lig_vec': lig_vec,
                'lig_ID': lig_ID}
    ret = (connect_ligands_to_core(lig_dict, core, min_dist) for core in input_cores)
    return list(chain.from_iterable(ret))


def multi_substitution(input_ligands, input_cores, n=1):
    """Attach ligands to cores; repeat this process n times."""
    ret = []
    mol_list = input_cores
    for _ in range(n):
        mol_list = substitution(input_ligands, mol_list)
        ret.append(mol_list)
    return list(chain.from_iterable(ret))



"""A module which adds the :class:`.FrozenSettings` class, an immutable counterpart to plams.Settings_.

.. _plams.Settings: https://www.scm.com/doc/plams/components/settings.html

Index
-----
.. currentmodule:: CAT.frozen_settings
.. autosummary::
    FrozenSettings

API
---
.. autoclass:: FrozenSettings
    :members:
    :private-members:
    :special-members:

"""  # noqa: E501

import copy
import textwrap
from typing import (NoReturn, Hashable)

from scm.plams import Settings

__all__ = ['FrozenSettings']


class FrozenSettings(Settings):
    """An inmutable subclass of plams.Settings_.

    .. _plams.Settings: https://www.scm.com/doc/plams/components/settings.html
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the construction of a :class:`FrozenSettings` instance."""
        dict.__init__(self, *args, **kwargs)
        cls = type(self)

        # Fill the FrozenSettings instance by means of the dict.__setitem__ method
        for key, value in self.items():
            if isinstance(value, dict):
                value_new = cls(value)
            elif isinstance(value, list):
                value_new = [cls(i) if isinstance(i, dict) else i for i in value]
            else:
                continue
            dict.__setitem__(self, key, value_new)

        # An attribute for caching the hash of this instance
        dict.__setattr__(self, '_hash', None)

    def __str__(self) -> str:
        """Return a string representation of this instance."""
        if not self:
            return f'{self.__class__.__name__}()'

        indent = 4 * ' '
        ret = super().__str__()[:-1]
        return f'{self.__class__.__name__}(\n{textwrap.indent(ret, indent)}\n)'

    __repr__ = __str__

    def __missing__(self, key: Hashable) -> 'FrozenSettings':
        """Return :data:`_frozen_settings`."""
        return _frozen_settings

    def __bool__(self) -> bool:
        """Implement :class:`bool(self)<bool>`."""
        return bool(len(self))

    @classmethod
    def _raise_exc(cls, *args, **kwargs) -> NoReturn:
        """Method unsupported; raise a :exc:`TypeError`."""
        raise TypeError(f"'{cls.__name__}' object does not support item assignment or deletion")

    __delitem__ = __setitem__ = set_nested = _raise_exc

    def flatten(self, flatten_list: bool = True) -> 'FrozenSettings':
        """Return a flattened copy of this instance.

        New keys are constructed by concatenating the (nested) keys of this instance into tuples.
        Opposite of the :meth:`.Settings.unflatten` method.
        If *flatten_list* is ``True``, all nested lists will be flattened as well. Dictionary keys are replaced with list indices in such case.

        """  # noqa: E501
        ret = super().flatten(flatten_list)
        return type(self)(ret)

    def unflatten(self, unflatten_list: bool = True) -> 'FrozenSettings':
        """Return a nested copy of this instance.

        New keys are constructed by expanding the keys of this instance (*e.g.* tuples) into new nested |Settings| instances.
        If *unflatten_list* is ``True``, integers will be interpretted as list indices and are used for creating nested lists.
        Opposite of the :meth:`.flatten` method.
        """  # noqa: E501
        ret = super().unflatten(unflatten_list)
        return type(self)(ret)

    def __hash__(self) -> int:
        """Return the hash of this instance; pull it from :attr:`._hash` if possible."""
        if self._hash is not None:
            return dict.__getattribute__(self, '_hash')

        flat = super().flatten()
        ret = 0
        for k, v in flat.items():
            ret ^= hash(k + (v, ))
        dict.__setattr__(self, '_hash', ret)
        return ret

    def copy(self, deep: bool = False) -> 'FrozenSettings':
        """Create a copy of this instance.

        The returned instance is a recursive copy, the **deep** parameter only affecting the
        function used for copying non-:class:`dict` values:
        :func:`copy.copy` if ``deep=False`` and  :func:`copy.deepcopy` if ``deep=True``.

        """
        copy_func = copy.deepcopy if deep else copy.copy
        ret = type(self)()
        for key, value in self.items():
            value_cp = copy_func(value)
            Settings.__setitem__(ret, key, value_cp)
        return ret

    def __copy__(self) -> 'FrozenSettings':
        """Create a shallow copy of this instance by calling :meth:`FrozenSettings.copy`."""
        return self.copy(deep=False)

    def __deepcopy__(self, memo=None) -> 'FrozenSettings':
        """Create a deep copy of this instance by calling :meth:`FrozenSettings.copy`."""
        return self.copy(deep=True)


_frozen_settings = FrozenSettings()



"""A module with various RDKit-based string-to-mol parser.

Index
-----
.. currentmodule:: CAT._mol_str_parsers
.. autosummary::
    str_to_rdmol
    FormatEnum

API
---
.. autofunction:: str_to_rdmol
.. autoclass:: FormatEnum

"""

# flake8: noqa: E501

import enum
import textwrap
from typing import Any, Callable

from nanoutils import PartialPrepend
from packaging.version import Version

try:
    import rdkit
    from rdkit import Chem
except ModuleNotFoundError:
    # Precaution against sphinx mocking raising a TypeError
    FLAGS = 0
    RDKIT_VERSION = Version("0.0.0")
else:
    FLAGS = Chem.SanitizeFlags.SANITIZE_ALL ^ Chem.SanitizeFlags.SANITIZE_ADJUSTHS
    RDKIT_VERSION = Version(rdkit.__version__)

__all__ = ["FormatEnum", "str_to_rdmol"]


def str_to_rdmol(
    string: str,
    parser: Callable[[str], Chem.Mol],
    kind: str = "string",
    **kwargs: Any,
) -> Chem.Mol:
    """Convert a SMILES string into an rdkit Mol; supports explicit hydrogens."""
    # RDKit tends to remove explicit hydrogens if SANITIZE_ADJUSTHS is enabled
    try:
        mol = parser(string, **kwargs)
        assert mol is not None
        if not kwargs.get("sanitize", True):
            Chem.SanitizeMol(mol, sanitizeOps=FLAGS)
    except Exception as ex:
        err_str = string if "\n" not in string else "\n" + textwrap.indent(string, 4 * " ")
        raise ValueError(f'Failed to parse the following {kind}: {err_str!r}') from ex
    return mol


class FormatEnum(enum.Enum):
    """An enum with various rdkit-based string-parsing options."""

    FASTA = PartialPrepend(str_to_rdmol, Chem.MolFromFASTA, "FASTA string", sanitize=False)
    HELM = PartialPrepend(str_to_rdmol, Chem.MolFromHELM, "HELM string", sanitize=False)
    INCHI = PartialPrepend(str_to_rdmol, Chem.MolFromInchi, "InChi string", sanitize=False, removeHs=False)
    MOL2 = PartialPrepend(str_to_rdmol, Chem.MolFromMol2Block, "Mol2 block", sanitize=False, removeHs=False)
    MOL2_FILE = PartialPrepend(str_to_rdmol, Chem.MolFromMol2File, "Mol2 file", sanitize=False, removeHs=False)
    MOL = PartialPrepend(str_to_rdmol, Chem.MolFromMolBlock, "Mol block", sanitize=False, removeHs=False)
    MOL_FILE = PartialPrepend(str_to_rdmol, Chem.MolFromMolFile, "Mol file", sanitize=False, removeHs=False)
    PDB = PartialPrepend(str_to_rdmol, Chem.MolFromPDBBlock, "PDB block", sanitize=False, removeHs=False)
    PDB_FILE = PartialPrepend(str_to_rdmol, Chem.MolFromPDBFile, "PDB file", sanitize=False, removeHs=False)
    SEQUENCE = PartialPrepend(str_to_rdmol, Chem.MolFromSequence, "sequence string", sanitize=False)
    SMARTS = PartialPrepend(str_to_rdmol, Chem.MolFromSmarts, "SMARTS string")
    SMILES = PartialPrepend(str_to_rdmol, Chem.MolFromSmiles, "SMILES string", sanitize=False)
    TPL = PartialPrepend(str_to_rdmol, Chem.MolFromTPLBlock, "TPL block", sanitize=False)
    TPL_FILE = PartialPrepend(str_to_rdmol, Chem.MolFromTPLFile, "TPL file", sanitize=False)
    if RDKIT_VERSION >= Version("2020.03.3"):
        SVG = PartialPrepend(str_to_rdmol, Chem.MolFromRDKitSVG, "SVG string", sanitize=False, removeHs=False)
    if RDKIT_VERSION >= Version("2020.09.1"):
        PNG = PartialPrepend(str_to_rdmol, Chem.MolFromPNGString, "PNG string")
        PNG_FILE = PartialPrepend(str_to_rdmol, Chem.MolFromPNGFile, "PNG file")



"""A module for parsing the ``ligand.anchor`` keyword."""

import re
import operator
from typing import Union, Tuple, Iterable, SupportsFloat

from rdkit import Chem
from scm.plams import Units, PT
from schema import Schema, Use, Optional
from typing_extensions import TypedDict, SupportsIndex

from ..utils import AnchorTup, KindEnum, FormatEnum
from ..attachment.ligand_anchoring import _smiles_to_rdmol, get_functional_groups

__all__ = ["parse_anchors"]


class _UnparsedAnchorDictBase(TypedDict):
    group: "str | SupportsIndex"
    anchor_idx: "SupportsIndex | Iterable[SupportsIndex]"


class _UnparsedAnchorDict(_UnparsedAnchorDictBase, total=False):
    remove: "None | SupportsIndex | Iterable[SupportsIndex]"
    angle_offset: "None | SupportsFloat | SupportsIndex | bytes | str"
    dihedral: "None | SupportsFloat | SupportsIndex | bytes | str"
    kind: "None | str | KindEnum"
    group_format: "None | str | FormatEnum"


class _AnchorDict(TypedDict):
    group: str
    group_idx: Tuple[int, ...]
    remove: "None | Tuple[int, ...]"
    kind: KindEnum
    angle_offset: "None | float"
    dihedral: "None | float"
    group_format: "FormatEnum"


def _parse_group_idx(item: "SupportsIndex | Iterable[SupportsIndex]") -> Tuple[int, ...]:
    """Parse the ``group_idx`` option."""
    try:
        return (operator.index(item),)
    except TypeError:
        pass

    ret = tuple(operator.index(i) for i in item)  # type: ignore[union-attr]
    n = len(ret) - len(set(ret))
    if n:
        raise ValueError(f"Found {n} duplicate elements")
    elif not ret:
        raise ValueError("Requires at least one element")
    return ret


def _parse_remove(
    item: "None | SupportsIndex | Iterable[SupportsIndex]"
) -> "None | Tuple[int, ...]":
    """Parse the ``remove`` option."""
    if item is None:
        return None
    return _parse_group_idx(item)


def _parse_kind(typ: "None | str | KindEnum") -> KindEnum:
    """Parse the ``kind`` option."""
    if typ is None:
        return KindEnum.FIRST
    elif isinstance(typ, KindEnum):
        return typ
    elif isinstance(typ, str):
        return KindEnum[typ.upper()]
    raise TypeError("`kind` expected None or a string")


def _parse_group_format(typ: "None | str | FormatEnum") -> FormatEnum:
    """Parse the ``group_format`` option."""
    if typ is None:
        return FormatEnum.SMILES
    elif isinstance(typ, FormatEnum):
        return typ
    elif isinstance(typ, str):
        return FormatEnum[typ.upper()]
    raise TypeError("`group_format` expected None or a string")


_UNIT_PATTERN = re.compile(r"([\.\_0-9]+)(\s+)?(\w+)?")


def _parse_angle_offset(
    offset: "None | SupportsFloat | SupportsIndex | bytes | str"
) -> "None | float":
    """Parse the ``angle_offset`` and ``dihedral`` options; convert the offset to radians."""
    if offset is None:
        return None
    elif not isinstance(offset, str):
        return Units.convert(float(offset), "deg", "rad")

    # match offsets such as `"5.5 degree"`
    match = _UNIT_PATTERN.match(offset)
    if match is None:
        raise ValueError(f"Invalid offset string: {offset!r}")

    offset, _, unit = match.groups()
    if unit is None:
        unit = "deg"
    return Units.convert(float(offset), unit, "rad")


def _parse_group(group: "str | SupportsIndex") -> str:
    """Parse the ``group`` option."""
    # Pre-process in case the group is passed as an atomic number
    try:
        atnum = operator.index(group)
    except TypeError:
        pass
    else:
        group = PT.get_symbol(atnum)

    # String parsing
    if not isinstance(group, str):
        raise TypeError("`group` expected a string or integer")
    elif group in DUMMY_SYMBOLS:
        return "*"
    else:
        return group


def _symbol_to_rdmol(symbol: str) -> Chem.Mol:
    """Helper function for converting atomic symbols to rdkit molecules."""
    atom = Chem.Atom(PT.get_atomic_number(symbol))
    mol = Chem.EditableMol(Chem.Mol())
    mol.AddAtom(atom)
    return mol.GetMol()


anchor_schema = Schema({
    "group": Use(_parse_group),
    "group_idx": Use(_parse_group_idx),
    Optional("remove", default=None): Use(_parse_remove),
    Optional("kind", default=KindEnum.FIRST): Use(_parse_kind),
    Optional("angle_offset", default=None): Use(_parse_angle_offset),
    Optional("dihedral", default=None): Use(_parse_angle_offset),
    Optional("group_format", default=FormatEnum.SMILES): Use(_parse_group_format),
})

#: A collection of symbols used for different kinds of dummy atoms.
DUMMY_SYMBOLS = frozenset(PT.dummysymbols)

#: All atom types that are not directly supported by the rdkit SMILES parser.
INVALID_SMILES_ATOMS = frozenset(
    PT.symtonum.keys() - DUMMY_SYMBOLS - {'B', 'Br', 'C', 'Cl', 'F', 'I', 'N', 'O', 'P', 'S'}
)


def parse_anchors(
    patterns: Union[
        None,
        SupportsIndex,
        str,
        Chem.Mol,
        AnchorTup,
        _UnparsedAnchorDict,
        "Iterable[str | SupportsIndex | Chem.Mol | AnchorTup | _UnparsedAnchorDict]",
    ] = None,
    split: bool = True,
    is_core: bool = False,
) -> Tuple[AnchorTup, ...]:
    """Parse the user-specified anchors."""
    if patterns is None:
        if is_core:
            raise TypeError("`anchor=None` is not supported for core anchors")
        patterns = get_functional_groups(None, split)
    elif isinstance(patterns, (Chem.Mol, str, dict, AnchorTup, SupportsIndex)):
        patterns = [patterns]

    ret = []
    for p in patterns:  # type: _UnparsedAnchorDict | str | Chem.Mol | SupportsIndex | AnchorTup
        if isinstance(p, AnchorTup):
            ret.append(p)
        elif isinstance(p, Chem.Mol):
            mol = p
            remove: "None | Tuple[int, ...]" = (
                None if not split else (list(mol.GetAtoms())[-1].GetIdx(),)
            )
            ret.append(AnchorTup(mol=mol, remove=remove))
        elif isinstance(p, dict):
            kwargs: _AnchorDict = anchor_schema.validate(p)

            group_idx = kwargs["group_idx"]
            remove = kwargs["remove"]
            angle_offset = kwargs["angle_offset"]
            dihedral = kwargs["dihedral"]
            group_parser = kwargs["group_format"].value
            group = kwargs["group"]
            if group in INVALID_SMILES_ATOMS:
                mol = _symbol_to_rdmol(group)
            else:
                mol = group_parser(group)

            # Dihedral and angle-offset options are not supported for core anchors
            if is_core:
                if dihedral is not None:
                    raise TypeError("`dihedral != None` is not supported for core anchors")
                elif angle_offset is not None:
                    raise TypeError("`angle_offset != None` is not supported for core anchors")
                elif kwargs["kind"] != KindEnum.FIRST:
                    raise NotImplementedError('`kind != "first"` is not yet supported')
            else:
                # Check that at least 3 atoms are available for `angle_offset`
                # (so a plane can be defined)
                if angle_offset is not None and len(group_idx) < 3:
                    raise ValueError("`group_idx` must contain at least 3 atoms when "
                                     "`angle_offset` is specified")

                # Check that at least 2 atoms are available for `dihedral`
                # (so the third dihedral-defining vector can be defined)
                if dihedral is not None and len(group_idx) < 2:
                    raise ValueError("`group_idx` must contain at least 2 atoms when "
                                     "`dihedral` is specified")

                # Check that `group_idx` and `remove` are disjoint
                # TODO: Investigate if this check can be removed
                if remove is not None and not set(group_idx).isdisjoint(remove):
                    raise ValueError("`group_idx` and `remove` must be disjoint")

            # Check that the indices in `group_idx` and `remove` are not out of bounds
            atom_count = len(mol.GetAtoms())
            if atom_count <= max(group_idx):
                raise IndexError(f"`group_idx` index {max(group_idx)} is out of bounds "
                                 f"for a `group` with {atom_count} atoms")
            elif remove is not None and atom_count <= max(remove):
                raise IndexError(f"`remove` index {max(remove)} is out of bounds "
                                 f"for a `group` with {atom_count} atoms")
            ret.append(AnchorTup(**kwargs, mol=mol))
        else:
            group = _parse_group(p)
            if group in INVALID_SMILES_ATOMS:
                mol = _symbol_to_rdmol(group)
            else:
                mol = _smiles_to_rdmol(group)
            remove = None if not split else (list(mol.GetAtoms())[-1].GetIdx(),)
            ret.append(AnchorTup(mol=mol, group=group, remove=remove))

    if is_core and len(ret) > 1:
        raise NotImplementedError("Cores with multiple anchor types aren't supported yet")
    return tuple(ret)



"""A context manager for temporary removing a set of atoms from a molecule.

Index
-----
.. currentmodule:: CAT.attachment.remove_atoms_cm
.. autosummary::
    RemoveAtoms

API
---
.. autoclass:: RemoveAtoms
    :members:

"""

import sys
from typing import Iterable, Union, Sequence, cast
from contextlib import AbstractContextManager

from scm.plams import Molecule, Atom, Bond, MoleculeError

if sys.version_info >= (3, 8):
    # dict.__reversed__() was added in Python 3.8
    from builtins import dict as OrderedDict  # noqa: N812
else:
    from collections import OrderedDict

__all__ = ['RemoveAtoms']


class RemoveAtoms(AbstractContextManager):
    """A context manager for temporary removing a set of atoms from a molecule.

    The *relative* ordering of the to-be removed atoms (and matching bonds),
    as specified in **atoms**, is preserved during the removal and reattachment process.
    Note that reattaching will (re-)append the removed atoms/bonds,
    a process which is thus likelly to affect the *absolute* ordering of
    atoms/bonds within the entire molecule.

    Examples
    --------
    .. code:: python

        >>> from scm.plams import Molecule, Atom, from_smiles

        >>> mol: Molecule = from_smiles('CO')
        >>> atom1: Atom = mol[1]
        >>> atom2: Atom = mol[2]

        >>> atom_set = {atom1, atom2}
        >>> with RemoveAtoms(mol, atom_set):
        ...     print(atom1 in mol, atom2 in mol)
        False False

        >>> print(atom1 in mol, atom2 in mol)
        True True

    Parameters
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule.
        See :attr:`RemoveAtoms.mol`.

    atoms : |plams.Atom| or |Iterable| [|plams.Atom|]
        A PLAMS atom or an iterable consisting of unique PLAMS atoms.
        All supplied atoms should belong to **mol**.
        See :attr:`RemoveAtoms.atoms`.

    Attributes
    ----------
    mol : |plams.Molecule|
        A PLAMS molecule.

    atoms : |Sequence| [|plams.Atom|]
        A sequence of PLAMS atoms belonging to :attr:`RemoveAtoms.mol`.
        Setting a value will convert it into a sequence of atoms.

    _bonds : |OrderedDict| [|plams.Bond|, ``None``]
        A ordered dictionary of PLAMS bonds connected to one or more atoms in
        :attr:`RemoveAtoms.atoms`.
        All values are ``None``, the dictionary serving as an improvised ``OrderedSet``.
        Set to ``None`` until :meth:`RemoveAtoms.__enter__` is called.

    """

    @property
    def atoms(self) -> Sequence[Atom]:
        """Get or set :attr:`RemoveAtoms.atoms`. Setting will convert the supplied value into a sequence of atoms."""  # noqa
        return self._atoms

    @atoms.setter
    def atoms(self, value: Union[Atom, Iterable[Atom]]) -> None:
        if isinstance(value, Atom):  # It's an atom
            self._atoms = (value,)
        elif not hasattr(value, '__reversed__'):  # It's an Iterator or Collection
            self._atoms = tuple(value)
        else:
            self._atoms = value  # It's a Sequence (probably)

    def __init__(self, mol: Molecule, atoms: Union[Atom, Iterable[Atom]]) -> None:
        """Initialize a :class:`RemoveAtoms` instance."""
        self.mol = mol
        self.atoms = cast(Sequence[Atom], atoms)
        self._bonds: 'OrderedDict[Bond, None]' = OrderedDict()  # An improvised "OrderedSet"

    def __enter__(self) -> None:
        """Enter the context manager; delete all atoms in :class:`RemoveAtoms.atoms`."""
        mol = self.mol
        bonds_set = self._bonds
        for atom in self.atoms:
            for bond in atom.bonds:
                bonds_set[bond] = None
            mol.delete_atom(atom)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager; reassign all atoms in :class:`RemoveAtoms.atoms`."""
        mol = self.mol
        for atom in reversed(self.atoms):
            mol.add_atom(atom)
        for bond in reversed(self._bonds):
            try:
                mol.add_bond(bond)
            except MoleculeError:
                pass  # One of the bonded atoms has been manually deleted by the user: skip it
        self._bonds = OrderedDict()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

# To update the package version number, edit CAT/__version__.py
version = {}
with open(os.path.join(here, 'CAT', '__version__.py')) as f:
    exec(f.read(), version)

with open('README.rst', encoding='utf-8') as readme_file:
    readme = readme_file.read()

setup(
    name='nlesc-CAT',
    version=version['__version__'],
    description=('A collection of tools designed for the automatic '
                 'construction of chemical compounds.'),
    long_description=readme + "\n\n",
    long_description_content_type='text/x-rst',
    author=['Bas van Beek', 'Jelena Belic'],
    author_email='b.f.van.beek@vu.nl',
    url='https://github.com/nlesc-nano/CAT',
    packages=[
        'CAT',
        'CAT.attachment',
        'CAT.data',
        'CAT.dye',
        'CAT.data.coskf',
        'CAT.data.templates',
        'CAT.data.CORES',
        'CAT.data.LIGANDS',
        'CAT.data_handling',
        'CAT.workflows'
    ],
    package_dir={'CAT': 'CAT'},
    package_data={
        'CAT': [
            'data/templates/*yaml',
            'data/coskf/*coskf',
            'data/coskf/misc/*coskf',
            'data/CORES/*xyz',
            'data/LIGANDS/*xyz',
            'workflows/workflow_yaml.yaml',
            'workflows/*.pyi',
            'py.typed'
        ]
    },
    entry_points={
        'console_scripts': ['init_cat=CAT.data_handling.entry_points:main']
    },
    include_package_data=True,
    license='GNU Lesser General Public License v3 or later',
    zip_safe=False,
    keywords=[
        'quantum-mechanics',
        'molecular-mechanics',
        'science',
        'chemistry',
        'python-3',
        'python-3-6',
        'python-3-7',
        'python-3-8',
        'python-3-9',
        'python-3-10',
        'automation',
        'scientific-workflows'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Typing :: Typed',
    ],
    test_suite='tests',
    python_requires='>=3.6',
    install_requires=[
        'Nano-Utils>=0.4.3',
        'numpy>=1.15.0',
        'scipy>=0.19.1',
        'pandas>=0.23.0',
        'pyyaml>=5.1',
        'schema>=0.7.2,!=0.7.5',
        'AssertionLib>=2.2.3',
        'plams>=1.5.1',
        'contextlib2>=0.6.0; python_version=="3.6"',
        'typing-extensions>=3.7.4.3',
        'qmflows>=0.11.0',
        'rdkit-pypi>=2018.03.1',
        'packaging>=1.16.8',
    ],
    setup_requires=[
        'pytest-runner',
    ],
    extras_require={
        'test': [
            'pytest>=5.4.0',
            'pytest-cov',
            'pytest-mock',
            'h5py>=2.7.0',
            'sphinx>=2.4; python_version>="3.7"',
            'sphinx_rtd_theme; python_version>="3.7"',
            'data-CAT>=0.7.0; python_version>="3.7"',
            'nano-CAT>=0.7.1; python_version>="3.7"',
            'auto-FOX>=0.10.0; python_version>="3.7"',
        ],
        'doc': [
            'sphinx>=2.4',
            'sphinx_rtd_theme',
            'data-CAT>=0.7.0',
            'nano-CAT>=0.7.1',
            'auto-FOX>=0.10.0',
        ],
    }
)



"""A module for optimizing rottion matrices.

Index
-----
.. currentmodule:: CAT.attachment.optimize_rotmat
.. autosummary::
    optimize_rotmat
API
---
.. autofunction:: optimize_rotmat

"""

from typing import Callable

import numpy as np
from scipy.optimize import minimize

from scm.plams import rotation_matrix, Atom

__all__ = ['optimize_rotmat']


def _minimize_func(vec1: np.ndarray, vec2: np.ndarray, xyz: np.ndarray, anchor: int) -> float:
    """The function whose output is to-be minimized by :func:`scipy.optimize.minimize`."""
    # Rotate and translate
    rotmat = rotation_matrix(vec1, vec2)
    xyz_new = xyz@rotmat
    xyz_new -= xyz_new[anchor]

    # Apply the cost function: e^(||yz||)
    yz = xyz_new[:, 1:]
    distance = np.linalg.norm(yz, axis=1)
    return np.exp(distance).sum()


def optimize_rotmat(mol: np.ndarray, anchor: int = 0,
                    func: Callable = _minimize_func) -> np.ndarray:
    r"""Find the rotation matrix for **xyz** that minimizes its deviation from the Cartesian X-axis.

    A set of vectors, :math:`v`, is constructed for all atoms in **xyz**,
    denoting their deviation from the Cartesian X-axis.
    Subsequently, the rotation matrix that minimizes :math:`\sum_{i}^{n} {e^{||v_{i}||}}`
    is returned.

    Parameters
    ----------
    mol : :class:`numpy.ndarray` [:class:`float`], shape :math:`(n,3)`
        An array-like object of Cartesian coordinates.
        *e.g.* :class:`list`, :class:`numpy.ndarray` or |plams.Molecule|.

    anchor : |plams.Atom| or :class:`int`
        The index (0-based) of the anchor atom in **xyz**.
        Used for defining the origin in **xyz** (*i.e.* :code:`xyz[i] == (0, 0, 0)`).
        Alternativelly, a PLAMS atom can be passed.

    Returns
    -------
    :class:`numpy.ndarray` [:class:`float`], shape :math:`(3,3)` or :math:`(3,)`
        An optimized rotation matrix.
        If ``as_vec=True``, instead return the initial vector
        used for constructing the rotation matrix.

    See Also
    --------
    :func:`minimize()<scipy.optimize.minimize>`
        Minimization of scalar function of one or more variables.

    """
    if hasattr(anchor, '__int__'):  # This encompasses both int and np.integer instances
        i = int(anchor)
    elif isinstance(anchor, Atom):
        i = anchor.mol.atoms.index(anchor)
    else:
        raise TypeError("The passed anchor is neither an 'int' nor 'Atom'; "
                        f"observed type: {repr(type(anchor))}")
    xyz = np.array(mol, dtype=float, ndmin=2, copy=False)

    # Create a first guess for the starting
    vec1 = xyz.mean(axis=0) - xyz[i]
    vec2 = np.array([1, 0, 0], dtype=float)
    with np.errstate(invalid='raise'):
        try:
            rotmat1 = rotation_matrix(vec1, vec2)
        except FloatingPointError:
            return np.eye(3)
    xyz_new = xyz@rotmat1.T

    # Optimize the trial vector; return the matching rotation matrix
    trial_vec = xyz_new.mean(axis=0) - xyz_new[i]
    output = minimize(_minimize_func, trial_vec, args=(vec2, xyz_new, i))
    rotmat2 = rotation_matrix(output.x, vec2)
    return (rotmat1.T@rotmat2.T).T



"""A module for managing exceptions and warnings in CAT."""

import warnings
from types import MappingProxyType
from typing import Mapping, Callable, NoReturn, Union, Type

from scm.plams import MoleculeError

from ..utils import MoleculeWarning as MoleculeWarning


__all__ = ['WARN_MAP']


class ValueWarning(Warning, ValueError):
    """A :exc:`Warning` subclass for :exc:`ValueError` related errors."""  # noqa


#: Map an :exc:`Exception` type to a :exc:`Warning` type.
CATEGORY_MAP: Mapping[Type[Exception], Type[Warning]] = MappingProxyType({
    MoleculeError: MoleculeWarning,
    ValueError: ValueWarning
})


def _warn(exc: Exception) -> None:
    """Perform a warning using **exc**.

    When possible, the warning category will be derived from the passed Exception type
    (see :data:`CATEGORY_MAP`).
    Will default to :exc:`RuntimeWarning` otherwise.

    """  # noqa
    warnings.warn(str(exc), stacklevel=2,
                  category=CATEGORY_MAP.get(type(exc), RuntimeWarning))


def _raise(exc: Exception) -> NoReturn:
    """Raise **exc**."""
    raise exc


def _ignore(exc: Exception) -> None:
    """Do nothing."""
    return None


#: Map a string to callable for either raising an :exc:`Exception`,
#: displaying a :exc:`Warning` or doing nothing.
WARN_MAP: Mapping[str, Callable[[Exception], Union[NoReturn, None]]] = MappingProxyType({
    'raise': _raise,
    'warn': _warn,
    'ignore': _ignore
})



"""A module with misc functions related to manipulating molecules and their geometry.

Index
-----
.. currentmodule:: CAT.mol_utils
.. autosummary::
    from_mol_other
    from_rdmol
    get_index
    separate_mod
    to_atnum
    to_symbol
    adf_connectivity
    fix_carboxyl
    fix_h

API
---
.. automethod:: from_mol_other
.. automethod:: from_rdmol
.. automethod:: get_index
.. automethod:: separate_mod
.. autofunction:: to_atnum
.. autofunction:: to_symbol
.. autofunction:: adf_connectivity
.. autofunction:: fix_carboxyl
.. autofunction:: fix_h

"""

from typing import Optional, Iterable, Union, Tuple, List

import numpy as np

from scm.plams import Molecule, Atom, Bond, MoleculeError, add_to_class, PTError
from scm.plams.tools.periodic_table import PeriodicTable
import scm.plams.interfaces.molecule.rdkit as molkit

from rdkit import Chem
from rdkit.Chem import rdMolTransforms

from ._mol_str_parser import FormatEnum

__all__ = ['adf_connectivity', 'fix_h', 'fix_carboxyl']


@add_to_class(Molecule)
def from_mol_other(self, mol: Molecule,
                   atom_subset: Optional[Iterable[Atom]] = None) -> None:
    """Update the Cartesian coordinates of this instance with those from another PLAMS molecule.

    Alternatively, update only a subset of atoms.

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule.

    atom_subset : |list|_ [|plams.Atom|_]
        Optional: A subset of atoms in **self**.

    """
    at_subset = atom_subset or self.atoms
    for at1, at2 in zip(at_subset, mol):
        at1.coords = at2.coords


@add_to_class(Molecule)
def from_rdmol(self, rdmol: Chem.Mol,
               atom_subset: Optional[Iterable[Atom]] = None) -> None:
    """Update the atomic coordinates of this instance with coordinates from an RDKit molecule.

    Alternatively, update only a subset of atoms.

    Parameters
    ----------
    rdmol : |rdkit.Chem.Mol|_
        An RDKit molecule.

    atom_subset : |list|_ [|plams.Atom|_]
        Optional: A subset of atoms in **self**.

    """
    at_subset = atom_subset or self.atoms
    conf = rdmol.GetConformer()
    for at1, at2 in zip(at_subset, rdmol.GetAtoms()):
        pos = conf.GetAtomPosition(at2.GetIdx())
        at1.coords = (pos.x, pos.y, pos.z)


@add_to_class(Molecule)
def get_index(self, value: Union[Atom, Bond]) -> Union[int, Tuple[int, int]]:
    """Return the first index of **value** within this instance.

    **value** expects an instance of either :class:`Atom` or :class:`Bond`.

    Note
    ----
    Following the convention addopted by PLAMS, the returned index/indices are 1-based rather
    than 0-based.

    Parameters
    ----------
    value : |plams.Atom|_ or |plams.Bond|_
        A PLAMS atom or bonds.

    Returns
    -------
    |int|_ or |tuple|_ [|int|_]
        An atomic index or (**value**: |plams.Atom|_) or
        a tuple of two atomic indices (**item**: |plams.Bond|_).

    Raises
    ------
    TypeError
        Raised if **value** is an instance of neither :class:`Atom` nor :class:`Bond`.

    MoleculeError
        Raised if the passed :class:`Atom` or :class:`Bond` is not in this instance.

    """
    if isinstance(value, Atom):
        if value not in self.atoms:
            raise MoleculeError("Passed atom, {repr(value)}, is not in this instance")
        return 1 + self.atoms.index(value)
    elif isinstance(value, Bond):
        if value not in self.bonds:
            raise MoleculeError(f"Passed bond, {repr(value)}, is not in this instance")
        at1, at2 = value
        return 1 + self.atoms.index(at1), 1 + self.atoms.index(at2)

    err = "item excepts an instance of 'Atom' or 'Bond'; observed type: '{}'"
    raise TypeError(err.format(value.__class__.__name__))


@add_to_class(Molecule)
def separate_mod(self) -> Tuple[Molecule]:
    """Modified PLAMS function: creates new molecules out of this instance rather than a copy of this instance. Atoms, bonds and properties are *not* copied.

    Separate the molecule into connected components.
    Returns is a list of new Molecule instrances (all atoms and bonds are disjoint with
    the original molecule).
    Each element of this list is identical to one connected component of the base molecule.
    A connected component is a subset of atoms such that there exists a path
    (along one or more bonds) between any two atoms.

    Returns
    -------
    |tuple|_ [|plams.Molecule|_]
        A list of molecules with atoms and bonds from **self**.

    """  # noqa: E501
    frags = ()
    for at in self:
        at._visited = False

    def dfs(v, mol):
        v._visited = True
        v.mol = mol
        for e in v.bonds:
            e.mol = mol
            u = e.other_end(v)
            if not u._visited:
                dfs(u, mol)

    for src in self.atoms:
        if not src._visited:
            m = Molecule()
            dfs(src, m)
            frags += (m,)
            m.properties = self.properties.copy()

    for at in self.atoms:
        del at._visited
        at.mol.atoms.append(at)
    for b in self.bonds:
        b.mol.bonds.append(b)

    return frags


@add_to_class(Molecule)
def round_coords(self, decimals: int = 3) -> None:
    """Round the Cartesian coordinates of this instance to a given precision in decimal digits.

    Performs an inplace update of all atoms in this instance.

    Parameters
    ----------
    decimals : int
        The desired precision in decimal digits.

    """
    xyz = self.as_array()
    xyz[:] = np.round(xyz, decimals=decimals)
    self.from_array(xyz)


def to_atnum(value: Union[str, int, Atom]) -> int:
    """Turn an atomic symbol into an atomic number.

    Parameters
    ----------
    value : |int|_ or |str|_
        An Atom, atomic symbol or number.

    Returns
    -------
    |int|_
        An atomic number.

    Raises
    ------
    TypeError
        Raised if **item** is an instance of neither :class:`str` nor :class:`int`.

    """
    if isinstance(value, str):
        try:
            return PeriodicTable.get_atomic_number(value)
        except PTError as ex:
            ex.args = (ex.args[0] + f': {repr(value)}',)
            raise ex
    elif isinstance(value, int):
        if value not in PeriodicTable.symtonum.values():
            raise PTError(f'trying to convert incorrect atomic number: {repr(value)}')
        return value

    try:
        return value.atnum
    except AttributeError as ex:
        raise TypeError("'value' expected an Atom or atomic symbol/number; "
                        f"observed type: '{value.__class__.__name__}'") from ex


def to_symbol(value: Union[str, int, Atom]) -> str:
    """Turn an atomic number into an atomic symbol.

    Parameters
    ----------
    value : |int|_ or |str|_
        An Atom, atomic symbol or number.

    Returns
    -------
    |int|_
        An atomic symbol.

    Raises
    ------
    TypeError
        Raised if **item** is an instance of neither :class:`str` nor :class:`int`.

    """
    if isinstance(value, int):
        try:
            return PeriodicTable.get_symbol(value)
        except PTError as ex:
            ex.args = (ex.args[0] + f': {repr(value)}',)
            raise ex
    elif isinstance(value, str):
        if value not in PeriodicTable.symtonum.keys():
            raise PTError(f'trying to convert incorrect atomic symbol: {repr(value)}')
        return value

    try:
        return value.symbol
    except AttributeError as ex:
        raise TypeError("'value' expected an Atom or atomic symbol/number; "
                        f"observed type: '{value.__class__.__name__}'") from ex


def adf_connectivity(mol: Molecule) -> List[str]:
    """Create an AMS-compatible connectivity list.

    Parameters
    ----------
    mol : |plams.Molecule|_
        A PLAMS molecule with :math:`n` bonds.

    Returns
    -------
    :math:`n` |list|_ [|str|_]
        An ADF-compatible connectivity list of :math:`n` bonds.

    """
    mol.set_atoms_id()

    # Create list of indices of all aromatic bonds
    try:
        rdmol = molkit.to_rdmol(mol)
    except Exception as ex:
        if type(ex) is ValueError or ex.__class__.__name__ == 'ArgumentError':
            # Plan B: ignore aromatic bonds
            bonds = [f'{bond.atom1.id} {bond.atom2.id} {bond.order:.1f}' for bond in mol.bonds]
            mol.unset_atoms_id()
            return bonds
        raise ex

    aromatic = [bond.GetIsAromatic() for bond in rdmol.GetBonds()]

    # Create a list of bond orders; aromatic bonds get a bond order of 1.5
    bond_orders = [(1.5 if ar else bond.order) for ar, bond in zip(aromatic, mol.bonds)]
    bonds = [f'{bond.atom1.id} {bond.atom2.id} {order:.1f}' for
             bond, order in zip(mol.bonds, bond_orders)]
    mol.unset_atoms_id()

    return bonds


_smiles_to_rdmol = FormatEnum.SMILES.value

#: A carboxylate
_CARBOXYLATE: Chem.Mol = _smiles_to_rdmol('[O-]C(C)=O')


def fix_carboxyl(mol: Molecule) -> None:
    """Resets carboxylate OCO angles if it is smaller than :math:`60` degrees.

    Performs an inplace update of **plams_mol**.

    Parameters
    ----------
    plams_mol : |plams.Molecule|_
        A PLAMS molecule.

    """
    rdmol = molkit.to_rdmol(mol)
    conf = rdmol.GetConformer()
    matches = rdmol.GetSubstructMatches(_CARBOXYLATE)

    if matches:
        get_angle = rdMolTransforms.GetAngleDeg
        set_angle = rdMolTransforms.SetAngleDeg
        for idx in matches:
            if get_angle(conf, idx[3], idx[1], idx[0]) < 60:
                set_angle(conf, idx[2], idx[1], idx[3], 180.0)
                set_angle(conf, idx[0], idx[1], idx[3], 120.0)
        mol.from_rdmol(rdmol)


def fix_h(mol: Molecule) -> None:
    """If a C=C-H angle is smaller than :math:`20` degrees, set it back to :math:`120` degrees.

    Performs an inplace update of **plams_mol**.

    Parameters
    ----------
    plams_mol : |plams.Molecule|_
        A PLAMS molecule.

    """
    h_list = [atom for atom in mol if atom.atnum == 1 and 2.0 in
              [bond.order for bond in mol.neighbors(atom)[0].bonds]]

    rdmol = molkit.to_rdmol(mol)
    conf = rdmol.GetConformer()
    get_idx = mol.atoms.index
    set_angle = rdMolTransforms.SetAngleDeg
    get_angle = rdMolTransforms.GetAngleDeg

    update = False
    for atom in h_list:
        at1 = atom  # Central atom
        at2 = mol.neighbors(at1)[0]  # Neighbours
        at3 = [atom for atom in mol.neighbors(at2) if atom != at1]  # Neighbours of neighbours

        # Create 2 sets of 3 atomic indices for defining angles: at1-at2=at3
        idx_tup1 = get_idx(at3[0]), get_idx(at2), get_idx(at1)
        idx_tup2 = get_idx(at3[1]), get_idx(at2), get_idx(at1)

        if get_angle(conf, *idx_tup1) <= 20.0:
            set_angle(conf, *idx_tup1, 120.0)
            update = True
        elif get_angle(conf, *idx_tup2) <= 20.0:
            set_angle(conf, *idx_tup2, 120.0)
            update = True

    if update:
        mol.from_rdmol(rdmol)



"""Modules designed for attaching ligands to cores."""

from .qd_opt import init_qd_opt
from .ligand_opt import init_ligand_opt
from .ligand_attach import init_qd_construction
from .ligand_anchoring import init_ligand_anchoring

from .edge_distance import edge_dist, plot_polyhedron
from .distribution import distribute_idx
from .distribution_brute import brute_uniform_idx
from .perp_surface import get_surface_vec, plot_vectors

__all__ = [
    'init_qd_opt',
    'init_ligand_opt',
    'init_qd_construction',
    'init_ligand_anchoring',

    'edge_dist', 'plot_polyhedron',
    'distribute_idx',
    'brute_uniform_idx',
    'get_surface_vec', 'plot_vectors'
]



"""Functions for creating distributions of atomic indices (*i.e.* core anchor atoms).

Index
-----
.. currentmodule:: CAT.distribution
.. autosummary::
    uniform_idx
    distribute_idx

API
---
.. autofunction:: uniform_idx
.. autofunction:: distribute_idx

"""

import sys
import reprlib
import functools
from types import MappingProxyType
from itertools import islice, takewhile
from collections import abc
from typing import (
    cast,
    Generator,
    Optional,
    Iterable,
    FrozenSet,
    Any,
    Union,
    Callable,
    Mapping,
    TYPE_CHECKING,
)

import numpy as np
from scipy.spatial.distance import cdist
from nanoutils import as_nd_array

from CAT.utils import cycle_accumulate
from CAT.attachment.edge_distance import edge_dist

__all__ = ['distribute_idx', 'uniform_idx']

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal

    _OpKind = Literal['min', 'max']
    _ModeKind = Literal['uniform', 'random', 'cluster']
else:
    ArrayLike = "numpy.typing.ArrayLike"
    _OpKind = "Literal['min', 'max']"
    _ModeKind = "Literal['uniform', 'random', 'cluster']"

#: A set of allowed values for the **mode** parameter in :func:`get_distribution`.
MODE_SET: FrozenSet[str] = frozenset({'uniform', 'random', 'cluster'})

#: Map the **operation** parameter in :func:`uniform_idx` to either
#: :func:`numpy.nanargmin` or :func:`numpy.nanargmax`.
_ArgFunc = Callable[[np.ndarray], np.intp]
OPERATION_MAPPING: Mapping[Union[str, Callable], _ArgFunc] = MappingProxyType({
    'min': np.nanargmin,
    min: np.nanargmin,
    np.min: np.nanargmin,
    'max': np.nanargmax,
    max: np.nanargmax,
    np.max: np.nanargmax,
})


def distribute_idx(
    core: ArrayLike,
    idx: Union[int, Iterable[int]],
    f: float,
    mode: _ModeKind = 'uniform',
    **kwargs: Any,
) -> np.ndarray:
    r"""Create a new distribution of atomic indices from **idx** of length :code:`f * len(idx)`.

    Parameters
    ----------
    core : array-like [:class:`float`], shape :math:`(m, 3)`
        A 2D array-like object (such as a :class:`Molecule` instance) consisting
        of Cartesian coordinates.

    idx : :class:`int` or :class:`Iterable<collections.abc.Iterable>` [:class:`int`], shape :math:`(i,)`
        An integer or iterable of unique integers representing the 0-based indices of
        all anchor atoms in **core**.

    f : :class:`float`
        A float obeying the following condition: :math:`0.0 < f \le 1.0`.
        Represents the fraction of **idx** that will be returned.

    mode : :class:`str`
        How the subset of to-be returned indices will be generated.
        Accepts one of the following values:

        * ``"random"``: A random distribution.
        * ``"uniform"``: A uniform distribution; the distance between each successive atom and
          all previous points is maximized.
        * ``"cluster"``: A clustered distribution; the distance between each successive atom and
          all previous points is minmized.

    \**kwargs : :data:`Any<typing.Any>`
        Further keyword arguments for the **mode**-specific functions.

    Returns
    -------
    :class:`numpy.ndarray` [:class:`int`], shape :math:`(f*i,)`
        A 1D array of atomic indices.
        If **idx** has :math:`i` elements,
        then the length of the returned list is equal to :math:`\max(1, f*i)`.

    See Also
    --------
    :func:`uniform_idx`
        Yield the column-indices of **dist** which yield a uniform or clustered distribution.

    :func:`cluster_idx`
        Return the column-indices of **dist** which yield a clustered distribution.

    """  # noqa
    # Convert **idx** into an array
    idx_ar = as_nd_array(idx, dtype=int)

    # Validate the input
    if mode not in MODE_SET:
        raise ValueError(f"Invalid `mode`: {mode!r}")
    elif not (0.0 < f <= 1.0):
        raise ValueError("'f' should be larger than 0.0 and smaller than or equal to 1.0")

    # Create an array of indices
    stop = max(1, int(round(f * len(idx_ar))))
    if mode in ('uniform', 'cluster'):
        xyz = np.array(core, dtype=np.float64, ndmin=2, copy=False)[idx_ar]
        dist = edge_dist(xyz) if kwargs.get('follow_edge', False) else cdist(xyz, xyz)
        operation: _OpKind = 'min' if mode == 'uniform' else 'max'

        generator1 = uniform_idx(
            dist,
            operation=operation,
            start=kwargs.get('start', None),
            cluster_size=kwargs.get('cluster_size', 1),
            randomness=kwargs.get('randomness', None),
            weight=kwargs.get('weight', lambda x: np.exp(-x))
        )
        generator2 = islice(generator1, stop)
        ret = idx_ar[np.fromiter(generator2, count=stop, dtype=np.intp)]
    elif mode == 'random':
        ret = np.random.permutation(idx)
    # Return a list of `p * len(idx)` atomic indices
    return ret[:stop]


def uniform_idx(
    dist: np.ndarray,
    operation: _OpKind = 'min',
    cluster_size: Union[int, Iterable[int]] = 1,
    start: Optional[int] = None,
    randomness: Optional[float] = None,
    weight: Callable[[np.ndarray], np.ndarray] = lambda x: np.exp(-x)
) -> Generator[int, None, None]:
    r"""Yield the column-indices of **dist** which yield a uniform or clustered distribution.

    Given the (symmetric) distance matrix :math:`\boldsymbol{D} \in \mathbb{R}^{n,n}` and
    the vector :math:`\boldsymbol{a} \in \mathbb{N}^{m}`
    (representing a subset of indices in :math:`D`),
    then the :math:`i`'th element :math:`a_{i}` is
    defined below. All elements of :math:`\boldsymbol{a}` are
    furthermore constrained to be unique.
    :math:`f(x)` is herein a, as of yet unspecified,
    function for weighting each individual distance.

    Following the convention used in python, the :math:`\boldsymbol{X}[0:3, 1:5]` notation is
    herein used to denote the submatrix created
    by intersecting rows :math:`0` up to (but not including) :math:`3` and
    columns :math:`1` up to (but not including) :math:`5`.

    .. math::

        \DeclareMathOperator*{\argmin}{\arg\!\min}
        a_{i} = \begin{cases}
            \argmin\limits_{k \in \mathbb{N}} \sum f \bigl( \boldsymbol{D}_{k,:} \bigr) &
            \text{if} & i=0 \\
            \argmin\limits_{k \in \mathbb{N}}
                \sum f \bigl( \boldsymbol{D}[k, \boldsymbol{a}[0:i]] \bigr) &
            \text{if} & i > 0
        \end{cases}

    Default weighting function: :math:`f(x) = e^{-x}`.

    The row in :math:`D` corresponding to :math:`a_{0}`
    can alternatively be specified by **start**.

    The :math:`\text{argmin}` operation can be exchanged for :math:`\text{argmax}` by setting
    **operation** to ``"max"``, thus yielding a clustered- rather than uniform-distribution.

    The **cluster_size** parameter allows for the creation of uniformly
    distributed clusters of size :math:`r`.
    Herein the vector of indices, :math:`\boldsymbol{a} \in \mathbb{N}^{m}` is
    for the purpose of book keeping reshaped
    into the matrix :math:`\boldsymbol{A} \in \mathbb{N}^{q, r} \; \text{with} \; q*r = m`.
    All elements of :math:`\boldsymbol{A}` are, again, constrained to be unique.

    .. math::

        \DeclareMathOperator*{\argmin}{\arg\!\min}
        A_{i,j} = \begin{cases}
            \argmin\limits_{k \in \mathbb{N}} \sum f \bigl( \boldsymbol{D}_{k,:} \bigr) &
            \text{if} & i=0; \; j=0 \\
            \argmin\limits_{k \in \mathbb{N}}
                \sum f \bigl( \boldsymbol{D}[k; \boldsymbol{A}[0:i, 0:r] \bigl) &
            \text{if} & i > 0; \; j = 0 \\
            \argmin\limits_{k \in \mathbb{N}}
                \dfrac{\sum f \bigl( \boldsymbol{D}[k, \boldsymbol{A}[0:i, 0:r] \bigr)}
                {\sum f \bigl( \boldsymbol{D}[k, \boldsymbol{A}[i, 0:j] \bigr)} &
            \text{if} & j > 0
        \end{cases}

    Examples
    --------
    .. code-block:: python

        >>> import numpy as np
        >>> from CAT.distribution import uniform_idx

        >>> dist: np.ndarray = np.random.rand(10, 10)

        >>> out1 = uniform_idx(dist)
        >>> idx_ar1 = np.fromiter(out1, dtype=np.intp)

        >>> out2 = uniform_idx(dist, operation="min")
        >>> out3 = uniform_idx(dist, cluster_size=5)
        >>> out4 = uniform_idx(dist, cluster_size=[1, 1, 1, 1, 2, 2, 4])
        >>> out5 = uniform_idx(dist, start=5)
        >>> out6 = uniform_idx(dist, randomness=0.75)
        >>> out7 = uniform_idx(dist, weight=lambda x: x**-1)

    Parameters
    ----------
    dist : :class:`numpy.ndarray` [:class:`float`], shape :math:`(n, n)`
        A symmetric 2D NumPy array (:math:`D_{i,j} = D_{j,i}`) representing the
        distance matrix :math:`D`.

    operation : :class:`str`
        Whether to use :func:`~numpy.argmin` or :func:`~numpy.argmax`.
        Accepted values are ``"min"`` and ``"max"``.

    cluster_size : :class:`int` or :class:`Iterable<collections.abc.Iterable>` [:class:`int`]
        An integer or iterable of integers representing the size of clusters.
        Used in conjunction with :code:`operation = "max"` for creating a uniform distribution
        of clusters.
        :code:`cluster_size = 1` is equivalent to a normal uniform distribution.

        Providing **cluster_size** as an iterable of integers will create clusters
        of varying, user-specified, sizes.
        For example, :code:`cluster_size = range(1, 4)` will continuesly create clusters
        of sizes 1, 2 and 3.
        The iteration process is repeated until all atoms represented by **dist** are exhausted.

    start : :class:`int`, optional
        The index of the starting row in **dist**.
        If ``None``, start in whichever row contains the global minimum
        (:math:`\DeclareMathOperator*{\argmin}{\arg\!\min} \argmin\limits_{k \in \mathbb{N}} ||\boldsymbol{D}_{k, :}||_{p}`) or maximum
        (:math:`\DeclareMathOperator*{\argmax}{\arg\!\max} \argmax\limits_{k \in \mathbb{N}} ||\boldsymbol{D}_{k, :}||_{p}`).
        See **operation**.

    randomness : :class:`float`, optional
        If not ``None``, represents the probability that a random index
        will be yielded rather than obeying **operation**.
        Should obey the following condition: :math:`0 \le randomness \le 1`.

    weight : :data:`Callable<typing.Callable>`
        A callable for applying weights to the distance; default: :math:`e^{-x}`.
        The callable should take an array as argument and return a new array,
        *e.g.* :func:`numpy.exp`.

    Yields
    ------
    :class:`int`
        Yield the column-indices specified in :math:`\boldsymbol{d}`.

    """  # noqa
    # Truncate and weight the distance matrix
    dist_w = np.array(dist, dtype=float, copy=True)
    if not np.diagonal(dist_w).any():
        np.fill_diagonal(dist_w, np.nan)
    dist_w = weight(dist_w)

    # Use either argmin or argmax
    try:
        arg_func = OPERATION_MAPPING[operation]
    except KeyError:
        raise ValueError(f"Invalid `operation`: {operation!r}") from None

    if start is None:
        start = cast(int, arg_func(np.nansum(dist_w, axis=1)))
    if randomness is not None:
        arg_func = _parse_randomness(randomness, arg_func, len(dist))

    # Yield the first index
    try:
        dist_w_1d: np.ndarray = dist_w[start].copy()
    except IndexError:
        if not hasattr(start, '__index__'):
            raise TypeError("'start' expected an integer or 'None'; observed type: "
                            f"'{start.__class__.__name__}'") from None
        raise ValueError("'start' is out of bounds") from None

    dist_w_1d[start] = np.nan
    yield start

    # Return a generator for yielding the remaining indices
    if cluster_size == 1:
        generator = _min_or_max(dist_w, dist_w_1d, arg_func)
    else:
        generator = _min_and_max(dist_w, dist_w_1d, arg_func, cluster_size)

    for i in generator:
        yield i


def _min_or_max(dist_sqr: np.ndarray, dist_1d_sqr: np.ndarray,
                arg_func: Callable[[np.ndarray], int]) -> Generator[int, None, None]:
    """Helper function for :func:`uniform_idx` if :code:`cluster_size == 1`."""
    for _ in range(len(dist_1d_sqr)-1):
        i = arg_func(dist_1d_sqr)
        dist_1d_sqr[i] = np.nan
        dist_1d_sqr += dist_sqr[i]
        yield i


def _min_and_max(dist_sqr: np.ndarray, dist_1d_sqr: np.ndarray,
                 arg_func: Callable[[np.ndarray], int],
                 cluster_size: Union[int, Iterable[int]] = 1) -> Generator[int, None, None]:
    """Helper function for :func:`uniform_idx` if :code:`cluster_size != 1`."""
    # Construct a boolean array denoting the start of new clusters
    bool_ar = np.zeros(len(dist_1d_sqr), dtype=bool)
    if isinstance(cluster_size, abc.Iterable):
        bool_indices = _parse_cluster_size(len(bool_ar), cluster_size)
        bool_ar[bool_indices] = True
    else:
        try:
            bool_ar[::cluster_size] = True
        except ValueError as ex:
            raise ValueError("'cluster_size' cannot be zero; oberved value: "
                             f"{reprlib.repr(cluster_size)}") from ex
        except TypeError as ex:
            raise TypeError("'cluster_size' expected a non-zero integer or iterable of integers; "
                            f"observed type: '{cluster_size.__class__.__name__}'") from ex
    bool_ar = bool_ar[1:]  # Skip the first element as it was already yielded in uniform_idx()

    dist_cluster = dist_1d_sqr.copy()
    for bool_ in bool_ar:
        if bool_:  # The start of a new cluster
            dist_1d_sqr += dist_cluster
            dist_cluster[:] = 0
            dist_1d = dist_1d_sqr
        else:  # The growth of a cluster
            dist_1d = dist_1d_sqr / dist_cluster

        j = arg_func(dist_1d)
        dist_1d_sqr[j] = np.nan
        dist_cluster += dist_sqr[j]
        yield j


def _random_arg_func(dist_1d: np.ndarray, arg_func: Callable[[np.ndarray], int],
                     threshold: float, idx: np.ndarray,
                     rand_func: Callable[[], float] = np.random.sample) -> int:
    """Return a random element from **idx** if :code:`threshold > rand_func()`, otherwise call :code:`arg_func(dist_1d)`.

    Elements in **dist_1d** which are `nan` will be used for masking **idx**.

    """  # noqa
    if threshold > rand_func():
        return np.random.choice(idx[~np.isnan(dist_1d)])
    return arg_func(dist_1d)


def _parse_cluster_size(ar_size: int, clusters: Iterable[int]) -> np.ndarray:
    """Return indices for all ``True`` values in the boolean array of :func:`_min_and_max`."""
    generator = takewhile(lambda x: x < ar_size, cycle_accumulate(clusters))
    try:
        return np.fromiter(generator, dtype=int)
    except TypeError as ex:
        raise TypeError("'cluster_size' expected a non-zero integer or iterable of integers; "
                        f"{ex}") from ex


def _parse_randomness(randomness: float, arg_func: Callable[[np.ndarray], int],
                      n: int) -> Callable[[np.ndarray], int]:
    """Modifiy **arg_func** such that there is a **randomness** chance to return a random index from the range **n**."""  # noqa
    try:
        assert (0 <= randomness <= 1)
    except TypeError as ex:
        raise TypeError("'randomness' expected a float larger than 0.0 and smaller than 1.0; "
                        f"observed type: '{randomness.__class__.__name__}'") from ex
    except AssertionError as ex:
        raise ValueError("'randomness' expected a float larger than 0.0 and smaller than 1.0; "
                         f"observed value: {reprlib.repr(randomness)}") from ex

    return functools.partial(_random_arg_func,
                             arg_func=arg_func,
                             threshold=randomness,
                             idx=np.arange(n))



"""A module with miscellaneous functions.

Index
-----
.. currentmodule:: CAT.utils
.. autosummary::
    type_to_string
    get_time
    check_sys_var
    dict_concatenate
    get_template
    KindEnum
    AnchorTup

API
---
.. autofunction:: type_to_string
.. autofunction:: get_time
.. autofunction:: check_sys_var
.. autofunction:: dict_concatenate
.. autofunction:: get_template
.. autoclass:: KindEnum
.. autoclass:: AnchorTup

"""

import os
import enum
import yaml
import inspect
import textwrap
import operator
import threading
import pkg_resources as pkg
from types import MappingProxyType, TracebackType
from shutil import rmtree
from pprint import pformat
from logging import Logger
from os.path import join, isdir, isfile, exists
from itertools import cycle, chain, repeat
from contextlib import redirect_stdout
from collections import abc, Counter
from typing import (
    Iterable, Union, TypeVar, Mapping, Type, Generator, Iterator, Optional,
    Any, NoReturn, Dict, overload, Callable, NamedTuple, Tuple,
)

import numpy as np
from scipy.spatial import cKDTree

from nanoutils import as_nd_array
from rdkit.Chem import Mol
from scm.plams.core.basejob import Job
from scm.plams.interfaces.thirdparty.orca import ORCAJob
from scm.plams import (
    config, Settings, Molecule, MoleculeError, PeriodicTable, init,
    from_smiles, AMSJob, ADFJob, Cp2kJob, DiracJob, GamessJob, CRSJob
)

from .logger import logger
from .mol_utils import to_atnum
from .gen_job_manager import GenJobManager
from ._mol_str_parser import FormatEnum

__all__ = [
    'JOB_MAP', 'check_sys_var', 'dict_concatenate', 'get_template',
    'cycle_accumulate', 'iter_repeat', 'get_nearest_neighbors',
    'log_traceback_locals', 'KindEnum', 'AnchorTup', 'AllignmentEnum',
    'AllignmentTup', 'FormatEnum',
]

JOB_MAP: Mapping[Type[Job], str] = MappingProxyType({
    ADFJob: 'adf',
    AMSJob: 'ams',
    DiracJob: 'dirac',
    Cp2kJob: 'cp2k',
    GamessJob: 'gamess',
    ORCAJob: 'orca',
    CRSJob: 'cosmo-rs'
})

T1 = TypeVar('T1')
T2 = TypeVar('T2')
KT = TypeVar('KT')
VT = TypeVar('VT')
P = TypeVar('P', str, bytes, os.PathLike)
Dtype = Union[type, str, np.dtype]


def type_to_string(job: Type[Job]) -> str:
    """Turn a :class:`type` instance into a :class:`str`."""
    try:
        return JOB_MAP[job]
    except KeyError:
        logger.error(f"No default settings available for type: '{job.__class__.__name__}'")
        return ''


def check_sys_var() -> None:
    """Validate all ADF environment variables.

    Raises
    ------
    EnvironmentError
        Raised if one or more of the following environment variables are absent:
        * ``'AMSBIN'``
        * ``'AMSHOME'``
        * ``'AMSRESOURCES'``
        * ``'SCMLICENSE'``

    """
    sys_var = ['AMSBIN', 'AMSHOME', 'AMSRESOURCES', 'SCMLICENSE']
    sys_var_exists = [item in os.environ for item in sys_var]
    for i, item in enumerate(sys_var_exists):
        if not item:
            logger.error(f"The environment variable {sys_var[i]!r} has not been set")

    if not all(sys_var_exists):
        raise EnvironmentError('One or more AMS environment variables have not been set, '
                               'aborting AMS job')


def dict_concatenate(dict_list: Iterable[Mapping[KT, VT]]) -> Dict[KT, VT]:
    """Concatenates a list of dictionaries."""
    ret: Dict[KT, VT] = {}
    for item in dict_list:
        ret.update(item)
    return ret


def get_template(template_name: str, from_cat_data: bool = True) -> Settings:
    """Grab a yaml template and return it as Settings object."""
    if from_cat_data:
        path = join('data/templates', template_name)
        xs = pkg.resource_string('CAT', path)
        return Settings(yaml.load(xs.decode(), Loader=yaml.FullLoader))
    with open(template_name, 'r') as file:
        return Settings(yaml.load(file, Loader=yaml.FullLoader))


@overload
def validate_path(path: None) -> str:
    ...
@overload   # noqa: E302
def validate_path(path: P) -> P:
    ...
def validate_path(path):  # noqa: E302
    """Validate a provided directory path.

    Parameters
    ----------
    path : str
        Optional: A path to a directory.
        Will default to the current working directory if ``None``.

    Results
    -------
    |str|_
        Returns either **path** or the current working directory.

    Raises
    ------
    FileNotFoundError
        Raised if **path** cannot be found.

    NotADirectoryError
        Raised if **path** is not a directory.

    """
    if path in {None, '.', ''}:
        return os.getcwd()
    elif isdir(path):
        return path

    if not exists(path):
        raise FileNotFoundError(f"{path!r} not found")
    elif isfile(path):
        raise NotADirectoryError(f"{path!r} is not a directory")


@overload
def validate_core_atom(atom: int) -> int:
    ...
@overload  # noqa: E302
def validate_core_atom(atom: str) -> Union[Molecule, int]:
    ...
def validate_core_atom(atom):  # noqa: E302
    """Parse and validate the ``["optional"]["qd"]["dissociate"]["core_atom"]`` argument."""
    # Potential atomic number or symbol
    if isinstance(atom, int) or atom in PeriodicTable.symtonum:
        return to_atnum(atom)

    # Potential SMILES string
    try:
        mol = from_smiles(atom)
    except Exception as ex:
        raise ValueError(f'Failed to recognize {atom!r} as a valid atomic number, '
                         'atomic symbol or SMILES string') from ex

    # Double check the SMILES string:
    charge_dict = {}
    for at in mol:
        charge = at.properties.charge
        try:
            charge_dict[charge] += 1
        except KeyError:
            charge_dict[charge] = 1
    if 0 in charge_dict:
        del charge_dict[0]

    # Only a single charged atom is allowed
    if len(charge_dict) > 1:
        charge_count = sum([v for v in charge_dict.values()])
        raise MoleculeError(f'The SMILES string {repr(atom)} contains more than one charged atom: '
                            f'charged atom count: {charge_count}')
    return mol


def parallel_init(path: Union[str, 'os.PathLike[str]'],
                  folder: Union[str, 'os.PathLike[str]'],
                  hashing: str = 'input') -> None:
    """Construct a workdir with a thread-safe name.

    Serves as a wrapper around :func:`plams.init<scm.plams.core.functions.init>`.

    Paramaters
    ----------
    path : str
        The path to the PLAMS workdir.
    folder : str
        The name of the PLAMS workdir.
    hashing : str
        The type of hashing used by the PLAMS :class:`JobManager`.
        Only ``"input"`` is considered an acceptable value here.

    """
    if hashing != 'input':
        raise ValueError(f"Invalid value: {hashing!r}")

    folder_ = f'{folder}.{threading.get_ident()}'
    path_abs = join(path, folder_)
    if isdir(path_abs):
        rmtree(path_abs)
    init(path=path, folder=folder_)

    # Not quite thread-safe, but as long as all threads assign a single value
    # there should be no problem
    config.log.file = 3
    config.log.stdout = 0
    return None


def restart_init(path: Union[str, 'os.PathLike[str]'],
                 folder: Union[str, 'os.PathLike[str]'],
                 hashing: str = 'input') -> None:
    """Wrapper around the plams.init_ function; used for importing one or more previous jobs.

    All pickled .dill files in **path**/**folder**/ will be loaded into the
    :class:`GenJobManager` instance initiated by :func:`init()<scm.plams.core.functions.init>`.

    .. _plams.init: https://www.scm.com/doc/plams/components/functions.html#scm.plams.core.functions.init

    Note
    ----
    Previous jobs are stored in a more generator-esque manner in :attr:`GenJobManager.hashes` and
    :class:`Job` instances are thus created on demand rather than
    permanently storing them in memory.

    Paramaters
    ----------
    path : str
        The path to the PLAMS workdir.

    folder : str
        The name of the PLAMS workdir.

    hashing : str
        Optional: The type of hashing used by the PLAMS :class:`JobManager`.
        Accepted values are: ``"input"``, ``"runscript"``, ``"input+runscript"`` and ``None``.

    """  # noqa
    # Create a job manager
    settings = Settings({'counter_len': 3, 'hashing': hashing, 'remove_empty_directories': True})
    manager = GenJobManager(settings, path, folder, hashing)

    # Change the default job manager
    with open(os.devnull, 'w') as f_, redirect_stdout(f_):
        init()
    if os.path.isdir(config.default_jobmanager.workdir):
        rmtree(config.default_jobmanager.workdir)
    config.default_jobmanager = manager
    config.log.file = 3
    config.log.stdout = 0

    workdir = manager.workdir
    if not isdir(workdir):  # workdir does not exist, dont bother trying to load previous jobs
        os.mkdir(workdir)
        return None

    # Update the default job manager with previous Jobs
    for f in os.listdir(workdir):
        job_dir = join(workdir, f)
        if not isdir(job_dir):  # Not a directory; move along
            continue

        hash_file = join(job_dir, f + '.hash')
        if isfile(hash_file):  # Update JobManager.hashes
            manager.load_job(hash_file)

        # Grab the job name
        try:
            name, _num = f.rsplit('.', 1)
            num = int(_num)
        except ValueError:  # Jobname is not appended with a number
            name = f
            num = 1

        # Update JobManager.names
        try:
            manager.names[name] = max(manager.names[name], num)
        except KeyError:
            manager.names[name] = num
    return None


@overload
def cycle_accumulate(iterable: Iterable[T1]) -> Generator[T1, None, None]:
    ...
@overload  # noqa: E302
def cycle_accumulate(iterable: Iterable[T1], start: T1) -> Generator[T1, None, None]:
    ...
def cycle_accumulate(iterable, start=0):  # noqa: E302
    """Accumulate and return elements from **iterable** until it is exhausted.

    Then repeat (and keep accumulating) the sequence indefinitely.
    The elements of **iterable** must have access to the :func:`__iadd__` method,
    *e.g.* :class:`float`, :class:`int` or :class:`str`.

    """
    ret = start
    for i in cycle(iterable):
        ret += i
        yield ret


def iter_repeat(iterable: Iterable[T1], times: int) -> Iterator[T1]:
    """Iterate over an iterable and apply :func:`itertools.repeat` to each element.

    Examples
    --------
    .. code:: python

        >>> from CAT.utils import iter_repeat

        >>> iterable = range(3)
        >>> n = 2
        >>> iterator = iter_repeat(iterable, n)
        >>> for i in iterator:
        ...     print(i)
        0
        0
        1
        1
        2
        2

    Parameters
    ----------
    iterable : :class:`Iterable<collections.abc.Iterable>`
        An iterable.

    times : :class:`int`
        The number of times each element should be repeated.

    Returns
    -------
    :class:`Iterator<collections.abc.Iterator>`
        An iterator that yields each element from **iterable** multiple **times**.

    """
    return chain.from_iterable(repeat(i, times) for i in iterable)


def get_nearest_neighbors(center: Union[Molecule, np.ndarray],
                          neighbor: Union[Molecule, np.ndarray],
                          k: Union[int, Iterable[int]],
                          distance_upper_bound: float = np.inf,
                          return_dist: bool = False, **kwargs: Any) -> np.ndarray:
    r"""Return the :math:`k` nearest-neighbors (from **neighbor**) for all user-specified atoms in **center**.

    The Euclidean distance is herein used for identifying nearest-neighbors.

    Warning
    -------
    Missing neighbors are denoted with :code:`len(center)` if
    an insufficient amount of neighbors can be identified.

    Parameters
    ----------
    center : array-like [:class:`float`], shape :math:`(n, 3)`
        A 2D array-like object with the Cartesian coordinates of all central atoms.

    neighbor : array-like [:class:`float`], shape :math:`(m, 3)`
        A 2D array-like object with the Cartesian coordinates of all
        (potentially) neighboring atoms.

    k : :class:`int` or :data:`Iterable<collections.abc.Iterable>` [:class:`int`]
        The number of to-be returned nearest neighbors.
        If :math:`k` is an iterable than it will return all nearest-neighbors specified in there.
        For example, :code:`[1, 3, 5]` will returned the first, third and fifth nearest-neighbors.
        Note that counting starts from 1.

    distance_upper_bound : :class:`float`
        Limit the nearest-neighbor search to neighbors within a certain radius with respect to
        each atom in **center**.

    return_dist : :class:`bool`
        If ``True``, return both the indices of the :math:`k` nearest neighbors and the respective
        :math:`(n, k)` distance matrix.

    \**kwargs : :data:`Any<typing.Any>`
        Further keyword arguments for SciPy's :class:`cKDTree<scipy.spatial.cKDTree>` class.

    Returns
    -------
    :class:`numpy.ndarray` [:class:`int`], shape :math:`(n, k)`
        A 2D array with indices of the  :math:`k` nearest neighbors.

    See Also
    --------
    :class:`cKDTree<scipy.spatial.cKDTree>`
        kd-tree for quick nearest-neighbor lookup.

    """  # noqa
    if center is neighbor:
        xyz1 = xyz2 = np.asarray(center)
    else:
        xyz1 = np.array(center, ndmin=2, copy=False)
        xyz2 = np.array(neighbor, ndmin=2, copy=False)

    if isinstance(k, abc.Iterable):
        k = as_nd_array(k, dtype=int)

    tree = cKDTree(xyz2, **kwargs)
    try:
        dist, idx = tree.query(xyz1, k=k, distance_upper_bound=distance_upper_bound)
    except ValueError as ex:
        _parse_value_error(ex, k)

    if idx.ndim == 1:  # Always return the indices as 2D array
        idx.shape += (1,)

    if return_dist:
        return dist, idx
    return idx


def _parse_value_error(ex: Exception, k: Any) -> NoReturn:
    """Post-process a :exc:`ValueError` raised by :func:`get_nearest_neighbors`."""
    if isinstance(k, abc.Iterable) and min(k) < 1:
        raise ValueError("All elements of 'k' must be larger than or equal to 1; "
                         f"observed minimum: {min(k)!r}") from ex
    elif hasattr(k, '__int__') and k < 1:
        raise ValueError("'k' must be larger than or equal to 1; "
                         f"observed value: {k!r}") from ex
    raise ex


class SetEnviron:
    """A reentrant, re-usable context manager for temporarily setting environment variables."""

    __slots__ = ('__weakref__', '_kwargs', '_kwargs_old')
    _kwargs: Mapping[str, str]
    _kwargs_old: Mapping[str, Optional[str]]

    def __init__(self, **kwargs: str) -> None:
        r"""Initialize the context manager.

        Parameters
        ----------
        \**kwargs : :class:`str`
            The to-be updated parameters.

        """
        self._kwargs = MappingProxyType(kwargs)
        self._kwargs_old = MappingProxyType({
            k: os.environ.get(k) for k in self._kwargs
        })

    def __enter__(self) -> None:
        """Enter the context manager."""
        os.environ.update(self._kwargs)

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> None:
        """Exit the context manager."""
        for k, v in self._kwargs_old.items():
            if v is None:
                # Use `pop` instead of `del` to ensure thread-safety
                os.environ.pop(k, None)
            else:
                os.environ[k] = v


def log_traceback_locals(logger: Logger, level: int = -1,
                         str_func: Callable[[object], str] = pformat) -> None:
    """Log all local variables at the specified traceback level.

    Parameters
    ----------
    logger : :class:`~logging.Logger`
        A logger for writing the local variables.
    level : :class:`int`
        The traceback level.
    str_func : :data:`Callable[[object], str]<typing.Callable>`
        The callable for creating the variables string representation.

    """
    try:
        local_dct = inspect.trace()[level].frame.f_locals
    except IndexError:
        i = operator.index(level)
        raise RuntimeError(f"No traceback was found at level {i}") from None

    for name, _value in local_dct.items():
        prefix = f"    {name}: {_value.__class__.__name__} = "
        n = len(prefix)
        value_str = textwrap.indent(str_func(_value), n * ' ')[n:].split('\n')
        value_str[0] = prefix + value_str[0]
        for v in value_str:
            logger.debug(v)


class KindEnum(enum.Enum):
    """An enum with different anchoring operation kinds (see :class:`AnchorTup`)."""

    FIRST = 0
    MEAN = 1
    MEAN_TRANSLATE = 2


class AllignmentEnum(enum.Enum):
    """An enum with different core vector orientations (see :class:`AllignmentTup`)."""

    SPHERE = 0
    SURFACE = 1


class AnchorTup(NamedTuple):
    """A named tuple with anchoring operation instructions."""

    mol: "None | Mol"
    group: "None | str" = None
    group_idx: Tuple[int, ...] = (0,)
    anchor_idx: Tuple[int, ...] = ()
    remove: "None | Tuple[int, ...]" = None
    kind: KindEnum = KindEnum.FIRST
    angle_offset: "None | float" = None
    dihedral: "None | float" = None
    group_format: FormatEnum = FormatEnum.SMILES


class AllignmentTup(NamedTuple):
    """A named tuple with core vector orientation options."""

    kind: AllignmentEnum
    invert: bool


class MoleculeWarning(Warning, MoleculeError):
    """A :exc:`Warning` subclass for :class:`~scm.plams.mol.molecule.Molecule` related errors."""


def get_formula(mol: Molecule) -> str:
    """Backport of the PLAMS <= 1.5.1 ``Molecule.get_formula`` method.

    The resulting atoms are reported in alphabetical order,
    contrary to the Hill system (that prioritizes ``CH`` pairs) utilized after 1.5.1.
    """
    dct = Counter(at.symbol for at in mol)
    return "".join(f"{at}{i}" for at, i in sorted(dct.items()))
